from abc import ABCMeta, abstractmethod
import numpy as np

import foxes.variables as FV
from foxes.utils import delta_wd


class ConvCrit(metaclass=ABCMeta):
    """
    Abstract base class for convergence criteria

    Attributes
    ----------
    name: str, optional
        The convergence criteria name

    :group: algorithms.iterative.models

    """

    def __init__(self, name=None):
        """
        Constructor.

        Parameters
        ----------
        name: str, optional
            The convergence criteria name

        """
        self.name = name if name is not None else type(self).__name__

    @abstractmethod
    def check_converged(self, algo, prev_results, results, verbosity=0):
        """
        Check convergence criteria.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        prev_results: xarray.Dataset
            The farm results of previous
            iteration, or None if first
        results: xarray.Dataset
            The farm results of current
            iteration
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        convergence: bool
            Convergence flag, true if converged

        """
        pass

    @abstractmethod
    def get_deltas(self):
        """
        Get the most recent evaluation deltas.

        Returns
        -------
        deltas: dict
            The most recent evaluation deltas

        """
        pass


class ConvCritList(ConvCrit):
    """
    A list of convergence criteria

    Attributes
    ----------
    crits: list of ConvCrit
        The criteria

    :group: algorithms.iterative.models

    """

    def __init__(self, crits=[], name=None):
        """
        Constructor.

        Parameters
        ----------
        crits: list of ConvCrit
            The criteria
        name: str, optional
            The convergence criteria name

        """
        super().__init__(name)
        self.crits = crits

    def add_crit(self, crit):
        """
        Add a convergence criterion

        Parameters
        ----------
        crit: ConvCrit
            The criterion

        """
        self.crits.append(crit)

    def check_converged(self, algo, prev_results, results, verbosity=0):
        """
        Check convergence criteria.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        prev_results: xarray.Dataset
            The farm results of previous
            iteration, or None if first
        results: xarray.Dataset
            The farm results of current
            iteration
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        convergence: bool
            Convergence flag, true if converged

        """
        self._failed = None
        for c in self.crits:
            if not c.check_converged(algo, prev_results, results, verbosity):
                self._failed = c
                return False

        return True

    def get_deltas(self):
        """
        Get the most recent evaluation deltas.

        Returns
        -------
        deltas: dict
            The most recent evaluation deltas

        """
        if self._failed is not None:
            return self._failed.get_deltas()
        return {}


class ConvVarDelta(ConvCrit):
    """
    Requires convergence of a selection of variables.

    Attributes
    ----------
    limits: dict
        The convergence limits. Keys: variables str,
        values: float values
    wd_vars: list of str
        The wind direction type variables (unit deg)

    :group: algorithms.iterative.models

    """

    def __init__(self, limits, wd_vars=None, name=None):
        """
        Constructor.

        Parameters
        ----------
        limits: dict
            The convergence limits. Keys: variables str,
            values: float values
        wd_vars: list of str, optional
            The wind direction type variables (unit deg)
        name: str, optional
            The convergence criteria name

        """
        super().__init__(name)
        self.limits = limits
        if wd_vars is None:
            self.wd_vars = [FV.WD, FV.AMB_WD, FV.YAW, FV.AMB_YAW]
        else:
            self.wd_vars = wd_vars

        self._deltas = {}

    def check_converged(self, algo, prev_results, results, verbosity=0):
        """
        Check convergence criteria.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        prev_results: xarray.Dataset
            The farm results of previous
            iteration, or None if first
        results: xarray.Dataset
            The farm results of current
            iteration
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        convergence: bool
            Convergence flag, true if converged

        """
        if prev_results is None:
            return False

        if verbosity > 0:
            print(f"\n{self.name}: Convergence check")
            L = max([len(v) for v in self.limits.keys()])

        ok = True
        self._deltas = {}
        for v, lim in self.limits.items():
            x0 = prev_results[v].to_numpy()
            x = results[v].to_numpy()
            if v in self.wd_vars:
                self._deltas[v] = np.max(np.abs(delta_wd(x0, x)))
            else:
                self._deltas[v] = np.max(np.abs(x - x0))
            check = self._deltas[v]
            ok = ok and (check <= lim)

            if verbosity > 0:
                r = "FAILED" if check > lim else "OK"
                print(f"  {v:<{L}}: delta = {check:.3e}, lim = {lim:.3e}  --  {r}")
            elif not ok:
                break

        return ok

    def get_deltas(self):
        """
        Get the most recent evaluation deltas.

        Returns
        -------
        deltas: dict
            The most recent evaluation deltas

        """
        return self._deltas


class DefaultConv(ConvVarDelta):
    """
    Default convergence criteria.

    :group: algorithms.iterative.models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__(
            {
                FV.REWS: 1e-6,
                FV.TI: 1e-7,
                FV.CT: 1e-7,
            }
        )



"""
Geometries in two dimensions.
"""

from .area_geometry import AreaGeometry, InvertedAreaGeometry
from .area_geometry import AreaUnion, AreaIntersection
from .polygon import ClosedPolygon
from .circle import Circle
from .half_plane import HalfPlane



import numpy as np

from foxes.core import AxialInductionModel


class BetzAxialInduction(AxialInductionModel):
    """
    The classic axial induction from 1D
    momentum theory

    Attributes
    ----------
    ct_max: float
        The maximal ct value

    :group: models.axial_induction

    """

    def __init__(self, ct_max=0.99999):
        """
        Constructor.

        Parameters
        ----------
        ct_max: float
            The maximal ct value

        """
        super().__init__()
        self.ct_max = ct_max

    def ct2a(self, ct):
        """
        Computes induction from ct

        Parameters
        ----------
        ct: numpy.ndarray or float
            The ct values

        Returns
        -------
        ct: numpy.ndarray or float
            The induction values

        """
        return 0.5 * (1 - np.sqrt(1 - np.minimum(ct, self.ct_max)))



import numpy as np

from .farm_data_model import FarmDataModelList, FarmDataModel
from .turbine_model import TurbineModel
from .turbine_type import TurbineType
import foxes.constants as FC


class FarmController(FarmDataModel):
    """
    Analyses selected turbine models and handles their call.

    Attributes
    ----------
    turbine_types: list of foxes.core.TurbineType
        The turbine type of each turbine
    turbine_model_names: list of str
        Names of all turbine models found in the farm
    pre_rotor_models: foxes.core.FarmDataModelList
        The turbine models with pre-rotor flag
    post_rotor_models: foxes.core.FarmDataModelList
        The turbine models without pre-rotor flag
    pars: dict
        Parameters for the turbine models, stored
        under their respecitve name

    :group: core

    """

    def __init__(self, pars={}):
        """
        Constructor.

        Parameters
        ----------
        pars: dict
            Parameters for the turbine models, stored
            under their respective name

        """
        super().__init__()

        self.turbine_types = None
        self.turbine_model_names = None
        self.pre_rotor_models = None
        self.post_rotor_models = None

        self.pars = pars

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [
            self.pre_rotor_models,
            self.post_rotor_models,
        ]

    def set_pars(self, model_name, init_pars, calc_pars, final_pars):
        """
        Set parameters for a turbine model

        Parameters
        ----------
        model_name: str
            Name of the model
        init_pars: dict
            Parameters for initialization
        calc_pars: dict
            Parameters for calculation
        final_pars: dict
            Parameters for finalization

        """
        self.pars[model_name] = {
            "init": init_pars,
            "calc": calc_pars,
            "final": final_pars,
        }

    def _analyze_models(self, algo, pre_rotor, models):
        """
        Helper function for model analysis
        """
        tmodels = []
        tmsels = []
        mnames = [[m.name for m in mlist] for mlist in models]
        tmis = np.zeros(algo.n_turbines, dtype=FC.ITYPE)
        news = True
        while news:
            news = False

            for ti, mlist in enumerate(models):
                if tmis[ti] < len(mlist):
                    mname = mnames[ti][tmis[ti]]
                    isnext = True
                    for tj, jnames in enumerate(mnames):
                        if (
                            tj != ti
                            and mname in jnames
                            and tmis[tj] < len(jnames)
                            and jnames[tmis[tj]] != mname
                        ):
                            isnext = False
                            break

                    if isnext:
                        m = models[ti][tmis[ti]]
                        tmodels.append(m)

                        tsel = np.zeros((algo.n_states, algo.n_turbines), dtype=bool)
                        for tj, jnames in enumerate(mnames):
                            mi = tmis[tj]
                            if mi < len(jnames) and jnames[mi] == mname:
                                ssel = algo.farm.turbines[tj].mstates_sel[mi]
                                tsel[:, tj] = True if ssel is None else ssel
                                tmis[tj] += 1
                        tmsels.append(tsel)

                        news = True
                        break

        if pre_rotor:
            self.pre_rotor_models = FarmDataModelList(tmodels)
            self.pre_rotor_models.name = f"{self.name}_prer"
            mtype = "pre-rotor"
        else:
            self.post_rotor_models = FarmDataModelList(tmodels)
            self.post_rotor_models.name = f"{self.name}_postr"
            mtype = "post-rotor"

        for ti, t in enumerate(algo.farm.turbines):
            if tmis[ti] != len(models[ti]):
                raise ValueError(
                    f"Turbine {ti}, {t.name}: Could not find turbine model order that includes all {mtype} turbine models, missing {t.models[tmis[ti]:]}"
                )

        return [m.name for m in tmodels], tmsels

    def collect_models(self, algo):
        """
        Analyze and gather turbine models, based on the
        turbines of the wind farm.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        """

        # check turbine models, and find turbine types and pre/post-rotor models:
        self.turbine_types = [None for t in algo.farm.turbines]
        prer_models = [[] for t in algo.farm.turbines]
        postr_models = [[] for t in algo.farm.turbines]
        for ti, t in enumerate(algo.farm.turbines):
            prer = None
            for mi, mname in enumerate(t.models):
                istype = False
                if mname in algo.mbook.turbine_types:
                    m = algo.mbook.turbine_types[mname]
                    if not isinstance(m, TurbineType):
                        raise TypeError(
                            f"Model {mname} type {type(m).__name__} is not derived from {TurbineType.__name__}"
                        )
                    models = [m]
                    istype = True
                elif mname in algo.mbook.turbine_models:
                    m = algo.mbook.turbine_models[mname]
                    models = m.models if isinstance(m, FarmDataModelList) else [m]
                    for mm in models:
                        if not isinstance(mm, TurbineModel):
                            raise TypeError(
                                f"Model {mname} type {type(mm).__name__} is not derived from {TurbineModel.__name__}"
                            )
                else:
                    raise KeyError(
                        f"Model {mname} not found in model book types or models"
                    )

                if istype:
                    if self.turbine_types[ti] is None:
                        self.turbine_types[ti] = m
                    else:
                        raise ValueError(
                            f"Turbine {ti}, {t.name}: Multiple turbine types found in self.turbine_models list, {self.turbine_types[ti].name} and {mname}"
                        )

                for m in models:
                    m.name = mname
                    if prer is None:
                        prer = m.pre_rotor
                    elif not prer and m.pre_rotor:
                        raise ValueError(
                            f"Turbine {ti}, {t.name}: Model is classified as pre-rotor, but following the post-rotor model '{t.models[mi-1]}'"
                        )
                    if m.pre_rotor:
                        prer_models[ti].append(m)
                    else:
                        postr_models[ti].append(m)

            if self.turbine_types[ti] is None:
                raise ValueError(
                    f"Turbine {ti}, {t.name}: Missing a turbine type model among models {t.models}"
                )

        # analyze models:
        mnames_pre, tmsels_pre = self._analyze_models(
            algo, pre_rotor=True, models=prer_models
        )
        mnames_post, tmsels_post = self._analyze_models(
            algo, pre_rotor=False, models=postr_models
        )
        tmsels = tmsels_pre + tmsels_post
        self._tmall = [np.all(t) for t in tmsels]
        self.turbine_model_names = mnames_pre + mnames_post
        if len(self.turbine_model_names):
            self._tmsels = np.stack(tmsels, axis=2)
        else:
            raise ValueError(f"Controller '{self.name}': No turbine model found.")

    def __get_pars(self, algo, models, ptype, mdata=None, downwind_index=None):
        """
        Private helper function for gathering model parameters.
        """
        pars = []
        for m in models:
            mi = self.turbine_model_names.index(m.name)
            if self._tmall[mi]:
                s = np.s_[:, :] if downwind_index is None else np.s_[:, downwind_index]
            else:
                if downwind_index is None:
                    s = mdata[FC.TMODEL_SELS][:, :, mi]
                else:
                    s = np.s_[
                        mdata[FC.TMODEL_SELS][:, downwind_index, mi], downwind_index
                    ]
            pars.append({"st_sel": s})
            if m.name in self.pars:
                pars[-1].update(self.pars[m.name][ptype])

        return pars

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        self.collect_models(algo)
        super().initialize(algo, verbosity)

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        idata = super().load_data(algo, verbosity)

        idata["coords"][FC.TMODELS] = self.turbine_model_names
        idata["data_vars"][FC.TMODEL_SELS] = (
            (FC.STATE, FC.TURBINE, FC.TMODELS),
            self._tmsels,
        )
        self._tmsels = None

        return idata

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return list(
            dict.fromkeys(
                self.pre_rotor_models.output_farm_vars(algo)
                + self.post_rotor_models.output_farm_vars(algo)
            )
        )

    def calculate(self, algo, mdata, fdata, pre_rotor, downwind_index=None):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        pre_rotor: bool
            Flag for running pre-rotor or post-rotor
            models
        downwind_index: int, optional
            The index in the downwind order

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        s = self.pre_rotor_models if pre_rotor else self.post_rotor_models
        pars = self.__get_pars(algo, s.models, "calc", mdata, downwind_index)
        res = s.calculate(algo, mdata, fdata, parameters=pars)
        return res

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 means silent

        """
        super().finalize(algo, verbosity)
        self.turbine_model_names = None



from math import sqrt
import foxes.models as fm
import foxes.variables as FV
from foxes.utils import FDict

from foxes.core import (
    PointDataModel,
    FarmDataModel,
    FarmController,
    RotorModel,
    TurbineType,
    TurbineModel,
    PartialWakesModel,
    WakeFrame,
    WakeSuperposition,
    WakeModel,
    AxialInductionModel,
    TurbineInductionModel,
    GroundModel,
)


class ModelBook:
    """
    Container for all kinds of models.

    Attributes
    ----------
    point_models: foxes.utils.FDict
        The point models. Keys: model name str,
        values: foxes.core.PointDataModel
    rotor_models: foxes.utils.FDict
        The rotor models. Keys: model name str,
        values: foxes.core.RotorModel
    turbine_types: foxes.utils.FDict
        The turbine type models. Keys: model name str,
        values: foxes.core.TurbineType
    turbine_models: foxes.utils.FDict
        The turbine models. Keys: model name str,
        values: foxes.core.TurbineModel
    farm_models: foxes.utils.FDict
        The farm models. Keys: model name str,
        values: foxes.core.FarmModel
    farm_controllers: foxes.utils.FDict
        The farm controllers. Keys: model name str,
        values: foxes.core.FarmController
    partial_wakes: foxes.utils.FDict
        The partial wakes. Keys: model name str,
        values: foxes.core.PartialWakeModel
    wake_frames: foxes.utils.FDict
        The wake frames. Keys: model name str,
        values: foxes.core.WakeFrame
    wake_superpositions: foxes.utils.FDict
        The wake superposition models. Keys: model name str,
        values: foxes.core.WakeSuperposition
    wake_models: foxes.utils.FDict
        The wake models. Keys: model name str,
        values: foxes.core.WakeModel
    induction_models: foxes.utils.FDict
        The induction models. Keys: model name str,
        values: foxes.core.AxialInductionModel
    ground_models: foxes.utils.FDict
        The ground models. Keys: model name str,
        values: foxes.core.GroundModel
    sources: foxes.utils.FDict
        All sources dict
    base_classes: foxes.utils.FDict
        The base classes for all model types

    :group: models

    """

    def __init__(self, Pct_file=None):
        """
        Constructor.

        Parameters
        ----------
        Pct_file: str, optional
            Path to power/ct curve file, for creation
            of default turbine type model
        """
        self.point_models = FDict(name="point_models")
        self.point_models["tke2ti"] = fm.point_models.TKE2TI()

        self.rotor_models = FDict(name="rotor_models")
        rvars = [FV.REWS, FV.REWS2, FV.REWS3, FV.TI, FV.RHO]
        self.rotor_models["centre"] = fm.rotor_models.CentreRotor(calc_vars=rvars)

        def _n2n(n2):
            n2 = float(n2)
            n = int(sqrt(n2))
            if n**2 != n2:
                raise Exception(
                    f"GridRotor factory: Value {n2} is not the square of an integer"
                )
            return n

        self.rotor_models.add_factory(
            fm.rotor_models.GridRotor,
            "grid<n2>",
            kwargs=dict(calc_vars=rvars, reduce=True),
            var2arg={"n2": "n"},
            n2=_n2n,
            hints={"n2": "(Number of points in square grid)"},
        )
        self.rotor_models.add_factory(
            fm.rotor_models.GridRotor,
            "grid<n2>_raw",
            kwargs=dict(calc_vars=rvars, reduce=False),
            var2arg={"n2": "n"},
            n2=_n2n,
            hints={"n2": "(Number of points in square grid)"},
        )
        self.rotor_models.add_factory(
            fm.rotor_models.LevelRotor,
            "level<n>",
            kwargs=dict(calc_vars=rvars, reduce=True),
            n=lambda x: int(x),
            hints={"n": "(Number of vertical levels)"},
        )
        self.rotor_models.add_factory(
            fm.rotor_models.LevelRotor,
            "level<n>_raw",
            kwargs=dict(calc_vars=rvars, reduce=False),
            n=lambda x: int(x),
            hints={"n": "(Number of vertical levels)"},
        )

        self.turbine_types = FDict(name="turbine_types")
        self.turbine_types["null_type"] = fm.turbine_types.NullType()
        self.turbine_types["NREL5MW"] = fm.turbine_types.PCtFile(
            "NREL-5MW-D126-H90.csv", rho=1.225
        )
        self.turbine_types["DTU10MW"] = fm.turbine_types.PCtFile(
            "DTU-10MW-D178d3-H119.csv", rho=1.225
        )
        self.turbine_types["IEA15MW"] = fm.turbine_types.PCtFile(
            "IEA-15MW-D240-H150.csv", rho=1.225
        )
        self.turbine_types["IWT7.5MW"] = fm.turbine_types.PCtFile(
            "IWT-7d5MW-D164-H100.csv", rho=1.225
        )
        if Pct_file is not None:
            self.turbine_types["Pct"] = fm.turbine_types.PCtFile(Pct_file)

        self.turbine_models = FDict(
            name="turbine_models",
            kTI=fm.turbine_models.kTI(),
            kTI_amb=fm.turbine_models.kTI(ti_var=FV.AMB_TI),
            thrust2ct=fm.turbine_models.Thrust2Ct(),
            PMask=fm.turbine_models.PowerMask(),
            yaw2yawm=fm.turbine_models.YAW2YAWM(),
            yawm2yaw=fm.turbine_models.YAWM2YAW(),
        )
        self.turbine_models.add_factory(
            fm.turbine_models.kTI,
            "kTI_<kTI>",
            kTI=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            hints={"kTI": "(Value, e.g. 004 for 0.04)"},
        )
        self.turbine_models.add_factory(
            fm.turbine_models.kTI,
            "kTI_amb_<kTI>",
            kwargs=dict(ti_var=FV.AMB_TI),
            kTI=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            hints={"kTI": "(Value, e.g. 004 for 0.04)"},
        )
        self.turbine_models.add_factory(
            fm.turbine_models.kTI,
            "kTI_<kTI>_<kb>",
            kTI=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            kb=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            hints={
                "kTI": "(Value, e.g. 004 for 0.04)",
                "kb": "(Value, e.g. 004 for 0.04)",
            },
        )
        self.turbine_models.add_factory(
            fm.turbine_models.kTI,
            "kTI_amb_<kTI>_<kb>",
            kwargs=dict(ti_var=FV.AMB_TI),
            kTI=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            kb=lambda x: float(f"0.{x[1:]}" if x[0] == "0" else float(x)),
            hints={
                "kTI": "(Value, e.g. 004 for 0.04)",
                "kb": "(Value, e.g. 004 for 0.04)",
            },
        )

        self.turbine_models["hubh_data"] = fm.turbine_models.RotorCentreCalc(
            {
                f"{FV.WD}_HH": FV.WD,
                f"{FV.WS}_HH": FV.WS,
                f"{FV.TI}_HH": FV.TI,
                f"{FV.RHO}_HH": FV.RHO,
            }
        )

        self.farm_models = FDict(
            name="farm_models",
            **{
                f"farm_{mname}": fm.farm_models.Turbine2FarmModel(m)
                for mname, m in self.turbine_models.items()
            },
        )

        self.farm_controllers = FDict(
            name="farm_controllers",
            basic_ctrl=fm.farm_controllers.BasicFarmController(),
        )

        self.partial_wakes = FDict(
            name="partial_wakes",
            rotor_points=fm.partial_wakes.RotorPoints(),
            top_hat=fm.partial_wakes.PartialTopHat(),
            centre=fm.partial_wakes.PartialCentre(),
        )
        self.partial_wakes.add_factory(
            fm.partial_wakes.PartialAxiwake,
            "axiwake<n>",
            n=lambda x: int(x),
            hints={"n": "(Number of evaluation points)"},
        )
        self.partial_wakes.add_factory(
            fm.partial_wakes.PartialGrid,
            "grid<n2>",
            var2arg={"n2": "n"},
            n2=_n2n,
            hints={"n2": "(Number of points in square grid)"},
        )

        self.wake_frames = FDict(
            name="wake_frames",
            rotor_wd=fm.wake_frames.RotorWD(var_wd=FV.WD),
            rotor_wd_farmo=fm.wake_frames.FarmOrder(),
        )
        self.wake_frames.add_k_factory(
            fm.wake_frames.YawedWakes,
            "yawed_[wake_k]",
        )
        self.wake_frames.add_factory(
            fm.wake_frames.Streamlines2D,
            "streamlines_<step>",
            step=lambda x: float(x),
            hints={"step": "(Step size in m)"},
        )
        self.wake_frames.add_factory(
            fm.wake_frames.Streamlines2D,
            "streamlines_<step>",
            step=lambda x: float(x),
            hints={"step": "(Step size in m)"},
        )

        self.wake_frames["timelines"] = fm.wake_frames.Timelines()

        def _todt(x):
            if x[-1] == "s":
                return float(x[:-1]) / 60
            elif x[-3:] == "min":
                return float(x[:-3])

        self.wake_frames.add_factory(
            fm.wake_frames.Timelines,
            "timelines_<dt>",
            dt=_todt,
            var2arg={"dt": "dt_min"},
            hints={"dt": "(Time step, e.g '10s', '1min' etc.)"},
        )
        self.wake_frames.add_factory(
            fm.wake_frames.SeqDynamicWakes,
            "seq_dyn_wakes_<dt>",
            dt=_todt,
            var2arg={"dt": "dt_min"},
            hints={"dt": "(Time step, e.g '10s', '1min' etc.)"},
        )

        self.wake_superpositions = FDict(
            name="wake_superpositions",
            ws_linear=fm.wake_superpositions.WSLinear(scale_amb=False),
            ws_linear_lim=fm.wake_superpositions.WSLinear(
                scale_amb=False, lim_low=1e-4
            ),
            ws_linear_amb=fm.wake_superpositions.WSLinear(scale_amb=True),
            ws_linear_amb_lim=fm.wake_superpositions.WSLinear(
                scale_amb=True, lim_low=1e-4
            ),
            ws_linear_loc=fm.wake_superpositions.WSLinearLocal(),
            ws_linear_loc_lim=fm.wake_superpositions.WSLinearLocal(lim_low=1e-4),
            ws_quadratic=fm.wake_superpositions.WSQuadratic(scale_amb=False),
            ws_quadratic_lim=fm.wake_superpositions.WSQuadratic(
                scale_amb=False, lim_low=1e-4
            ),
            ws_quadratic_amb=fm.wake_superpositions.WSQuadratic(scale_amb=True),
            ws_quadratic_amb_lim=fm.wake_superpositions.WSQuadratic(
                scale_amb=True, lim_low=1e-4
            ),
            ws_quadratic_loc=fm.wake_superpositions.WSQuadraticLocal(),
            ws_quadratic_loc_lim=fm.wake_superpositions.WSQuadraticLocal(lim_low=1e-4),
            ws_cubic=fm.wake_superpositions.WSPow(pow=3, scale_amb=False),
            ws_cubic_amb=fm.wake_superpositions.WSPow(pow=3, scale_amb=True),
            ws_cubic_loc=fm.wake_superpositions.WSPowLocal(pow=3),
            ws_cubic_loc_lim=fm.wake_superpositions.WSPowLocal(pow=3, lim_low=1e-4),
            ws_quartic=fm.wake_superpositions.WSPow(pow=4, scale_amb=False),
            ws_quartic_amb=fm.wake_superpositions.WSPow(pow=4, scale_amb=True),
            ws_quartic_loc=fm.wake_superpositions.WSPowLocal(pow=4),
            ws_quartic_loc_lim=fm.wake_superpositions.WSPowLocal(pow=4, lim_low=1e-4),
            ws_max=fm.wake_superpositions.WSMax(scale_amb=False),
            ws_max_amb=fm.wake_superpositions.WSMax(scale_amb=True),
            ws_max_loc=fm.wake_superpositions.WSMaxLocal(),
            ws_max_loc_lim=fm.wake_superpositions.WSMaxLocal(lim_low=1e-4),
            ws_product=fm.wake_superpositions.WSProduct(),
            ws_product_lim=fm.wake_superpositions.WSProduct(lim_low=1e-4),
            ti_linear=fm.wake_superpositions.TILinear(superp_to_amb="quadratic"),
            ti_quadratic=fm.wake_superpositions.TIQuadratic(superp_to_amb="quadratic"),
            ti_cubic=fm.wake_superpositions.TIPow(pow=3, superp_to_amb="quadratic"),
            ti_quartic=fm.wake_superpositions.TIPow(pow=4, superp_to_amb="quadratic"),
            ti_max=fm.wake_superpositions.TIMax(superp_to_amb="quadratic"),
        )

        self.axial_induction = FDict(name="induction_models")
        self.axial_induction["Betz"] = fm.axial_induction.BetzAxialInduction()
        self.axial_induction["Madsen"] = fm.axial_induction.MadsenAxialInduction()

        self.wake_models = FDict(name="wake_models")

        self.wake_models.add_k_factory(
            fm.wake_models.wind.JensenWake,
            "Jensen_<superposition>_[wake_k]",
            kwargs=dict(induction="Betz"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )

        self.wake_models.add_k_factory(
            fm.wake_models.wind.Bastankhah2014,
            "Bastankhah2014_<superposition>_[wake_k]",
            kwargs=dict(sbeta_factor=0.2, induction="Madsen"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.wind.Bastankhah2014,
            "Bastankhah2014B_<superposition>_[wake_k]",
            kwargs=dict(sbeta_factor=0.2, induction="Betz"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.wind.Bastankhah2014,
            "Bastankhah025_<superposition>_[wake_k]",
            kwargs=dict(sbeta_factor=0.25, induction="Madsen"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.wind.Bastankhah2014,
            "Bastankhah025B_<superposition>_[wake_k]",
            kwargs=dict(sbeta_factor=0.25, induction="Betz"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )

        self.wake_models.add_k_factory(
            fm.wake_models.wind.Bastankhah2016,
            "Bastankhah2016_<superposition>_[wake_k]",
            kwargs=dict(induction="Madsen"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.wind.Bastankhah2016,
            "Bastankhah2016B_<superposition>_[wake_k]",
            kwargs=dict(induction="Betz"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )

        self.wake_models.add_k_factory(
            fm.wake_models.wind.TurbOParkWake,
            "TurbOPark_<superposition>_[wake_k]",
            kwargs=dict(induction="Madsen"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.wind.TurbOParkWake,
            "TurbOParkB_<superposition>_[wake_k]",
            kwargs=dict(induction="Betz"),
            superposition=lambda s: f"ws_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ws_linear)"},
        )

        self.wake_models.add_k_factory(
            fm.wake_models.wind.TurbOParkWakeIX,
            "TurbOParkIX_<superposition>_[wake_k]_dx<dx>",
            superposition=lambda s: f"ws_{s}",
            dx=lambda x: float(x),
            hints={
                "superposition": "(Superposition, e.g. linear for ws_linear)",
                "dx": "(Integration step in m)",
            },
        )

        self.wake_models.add_k_factory(
            fm.wake_models.ti.CrespoHernandezTIWake,
            "CrespoHernandez_<superposition>_[wake_k]",
            kwargs=dict(use_ambti=False),
            superposition=lambda s: f"ti_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ti_linear)"},
        )

        self.wake_models.add_factory(
            fm.wake_models.ti.IECTIWake,
            "IECTI2005_<superposition>",
            kwargs=dict(iec_type="2005"),
            superposition=lambda s: f"ti_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ti_linear)"},
        )

        self.wake_models.add_factory(
            fm.wake_models.ti.IECTIWake,
            "IECTI2019_<superposition>",
            kwargs=dict(iec_type="2019"),
            superposition=lambda s: f"ti_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ti_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.ti.IECTIWake,
            "IECTI2019k_<superposition>_[wake_k]",
            kwargs=dict(iec_type="2019", opening_angle=None),
            superposition=lambda s: f"ti_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ti_linear)"},
        )

        self.wake_models.add_factory(
            fm.wake_models.ti.IECTIWake,
            "IECTI2005_<superposition>",
            kwargs=dict(iec_type="2005"),
            superposition=lambda s: f"ti_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ti_linear)"},
        )
        self.wake_models.add_k_factory(
            fm.wake_models.ti.IECTIWake,
            "IECTI2005k_<superposition>_[wake_k]",
            kwargs=dict(iec_type="2005", opening_angle=None),
            superposition=lambda s: f"ti_{s}",
            hints={"superposition": "(Superposition, e.g. linear for ti_linear)"},
        )

        self.wake_models[f"RHB"] = fm.wake_models.induction.RankineHalfBody()
        self.wake_models[f"Rathmann"] = fm.wake_models.induction.Rathmann()
        self.wake_models[f"SelfSimilar"] = fm.wake_models.induction.SelfSimilar()
        self.wake_models[f"SelfSimilar2020"] = (
            fm.wake_models.induction.SelfSimilar2020()
        )
        self.wake_models[f"VortexSheet"] = fm.wake_models.induction.VortexSheet()

        self.ground_models = FDict(name="ground_models")
        self.ground_models["no_ground"] = fm.ground_models.NoGround()
        self.ground_models["ground_mirror"] = fm.ground_models.GroundMirror()
        self.ground_models.add_factory(
            fm.ground_models.WakeMirror,
            "blh_mirror_h<height>",
            var2arg={"height": "heights"},
            height=lambda h: [0.0, float(h)],
            hints={"height": "(Boundary layer wake reflection height)"},
        )

        self.sources = FDict(
            name="sources",
            point_models=self.point_models,
            rotor_models=self.rotor_models,
            turbine_types=self.turbine_types,
            turbine_models=self.turbine_models,
            farm_models=self.farm_models,
            farm_controllers=self.farm_controllers,
            partial_wakes=self.partial_wakes,
            wake_frames=self.wake_frames,
            wake_superpositions=self.wake_superpositions,
            wake_models=self.wake_models,
            axial_induction=self.axial_induction,
            ground_models=self.ground_models,
        )
        self.base_classes = FDict(
            name="base_classes",
            point_models=PointDataModel,
            rotor_models=RotorModel,
            turbine_types=TurbineType,
            turbine_models=TurbineModel,
            farm_models=FarmDataModel,
            farm_controllers=FarmController,
            partial_wakes=PartialWakesModel,
            wake_frames=WakeFrame,
            wake_superpositions=WakeSuperposition,
            wake_models=WakeModel,
            axial_induction=AxialInductionModel,
            ground_models=GroundModel,
        )

        for s in self.sources.values():
            for k, m in s.items():
                m.name = k

    def __getitem__(self, key):
        return self.sources.__getitem__(key)

    def print_toc(self, subset=None, search=None):
        """
        Print the contents.

        Parameters
        ----------
        subset: list of str, optional
            Selection of model types
        search:  str, optional
            String that has to be part of the model name

        """

        for k in sorted(list(self.sources.keys())):
            ms = self.sources[k]
            if subset is None or k in subset:
                print(k)
                print("-" * len(k))
                if len(ms):
                    for mname in sorted(list(ms.keys())):
                        if search is None or search in mname:
                            print(f"{mname}: {ms[mname]}")
                    if isinstance(ms, FDict):
                        for f in ms.factories:
                            if search is None or search in f.name_template:
                                print()
                                print(f)
                else:
                    print("(none)")
                print()

    def get(self, model_type, name, class_name=None, *args, **kwargs):
        """
        Gets a model object.

        If not found, dynamically creates it (given the class name)

        Parameters
        ----------
        model_type: str
            The model type
        name: str
            The model name
        class_name: str, optinal
            Name of the model class
        args: tuple, optional
            Arguments for the model class
        kwargs: dict, optional
            Arguments for the model class

        Returns
        -------
        model: mclass
            The model object

        """
        if name not in self.sources[model_type]:
            if class_name is None:
                raise KeyError(
                    f"Model '{name}' of type '{model_type}' not found in model book. Available: {sorted(list(self.sources[model_type].keys()))}"
                )
            bclass = self.base_classes[model_type]
            self.sources[model_type][name] = bclass.new(class_name, *args, **kwargs)
        return self.sources[model_type][name]

    def default_partial_wakes(self, wake_model):
        """
        Gets a default partial wakes model name
        for a given wake model

        Parameters
        ----------
        wake_model: foxes.core.WakeModel
            The wake model

        Returns
        -------
        pwake: str
            The partial wake model name

        """
        if isinstance(wake_model, TurbineInductionModel):
            return "grid9"
        elif isinstance(wake_model, fm.wake_models.TopHatWakeModel):
            return "top_hat"
        elif isinstance(wake_model, fm.wake_models.AxisymmetricWakeModel):
            return "axiwake6"
        elif isinstance(wake_model, fm.wake_models.DistSlicedWakeModel):
            return "grid9"
        else:
            raise TypeError(
                f"No default partial wakes model defined for wake model type '{type(wake_model).__name__}'"
            )

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        for ms in self.sources.values():
            if isinstance(ms, FDict):
                for m in ms.values():
                    if m.initialized:
                        m.finalize(algo, verbosity)



"""
Wake models.
"""

from .dist_sliced import DistSlicedWakeModel
from .axisymmetric import AxisymmetricWakeModel
from .top_hat import TopHatWakeModel
from .gaussian import GaussianWakeModel

from . import wind
from . import ti
from . import induction



import numpy as np

from foxes.core import FarmDataModel, TData
import foxes.variables as FV
import foxes.constants as FC


class InitFarmData(FarmDataModel):
    """
    Sets basic turbine data and applies downwind order

    :group: algorithms.downwind.models

    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__(pre_rotor=True)

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [
            FV.X,
            FV.Y,
            FV.H,
            FV.D,
            FV.WD,
            FV.YAW,
            FV.ORDER,
            FV.WEIGHT,
            FV.ORDER_SSEL,
            FV.ORDER_INV,
        ]

    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # prepare:
        n_states = fdata.n_states
        n_turbines = algo.n_turbines

        # define FV.TXYH as vector [X, Y, H]:
        fdata[FV.TXYH] = np.full((n_states, n_turbines, 3), np.nan, dtype=FC.DTYPE)
        fdata.dims[FV.TXYH] = (FC.STATE, FC.TURBINE, FC.XYH)
        for i, v in enumerate([FV.X, FV.Y, FV.H]):
            fdata[v] = fdata[FV.TXYH][..., i]
            fdata.dims[v] = (FC.STATE, FC.TURBINE)

        # set X, Y, H, D:
        fdata[FV.D] = np.zeros((n_states, n_turbines), dtype=FC.DTYPE)
        for ti, t in enumerate(algo.farm.turbines):

            if len(t.xy.shape) == 1:
                fdata[FV.TXYH][:, ti, :2] = t.xy[None, :]
            else:
                i0 = fdata.states_i0()
                s = np.s_[i0 : i0 + fdata.n_states]
                fdata[FV.TXYH][:, ti, :2] = t.xy[s]

            H = t.H
            if H is None:
                H = algo.farm_controller.turbine_types[ti].H
            fdata[FV.TXYH][:, ti, 2] = H

            D = t.D
            if D is None:
                D = algo.farm_controller.turbine_types[ti].D
            fdata[FV.D][:, ti] = D

        # calc WD and YAW at rotor centres:
        tdata = TData.from_points(points=fdata[FV.TXYH])
        sres = algo.states.calculate(algo, mdata, fdata, tdata)
        fdata[FV.WD] = sres[FV.WD][:, :, 0]
        del tdata, sres

        # calculate and inverse:
        order = algo.wake_frame.calc_order(algo, mdata, fdata)
        ssel = np.zeros_like(order)
        ssel[:] = np.arange(n_states)[:, None]
        fdata[FV.ORDER] = order
        fdata[FV.ORDER_SSEL] = ssel
        fdata[FV.ORDER_INV] = np.zeros_like(order)
        fdata[FV.ORDER_INV][ssel, order] = np.arange(n_turbines)[None, :]

        # apply downwind order to all data:
        fdata[FV.TXYH] = fdata[FV.TXYH][ssel, order]
        for i, v in enumerate([FV.X, FV.Y, FV.H]):
            fdata[v] = fdata[FV.TXYH][..., i]
        for v in [FV.D, FV.WD, FV.WEIGHT]:
            if np.any(fdata[v] != fdata[v][0, 0, None, None]):
                fdata[v] = fdata[v][ssel, order]
        fdata[FV.YAW] = fdata[FV.WD].copy()
        for k in mdata.keys():
            if tuple(mdata.dims[k][:2]) == (FC.STATE, FC.TURBINE) and np.any(
                mdata[k] != mdata[k][0, 0, None, None]
            ):
                mdata[k] = mdata[k][ssel, order]

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



import numpy as np

from foxes.core import WakeFrame, WakeK, TData
from foxes.models.wake_models.wind.bastankhah16 import (
    Bastankhah2016Model,
    Bastankhah2016,
)
import foxes.variables as FV
import foxes.constants as FC
from .rotor_wd import RotorWD


class YawedWakes(WakeFrame):
    """
    Bend the wakes for yawed turbines, based on the
    Bastankhah 2016 wake model

    Notes
    -----
    Reference:
    "Experimental and theoretical study of wind turbine wakes in yawed conditions"
    Majid Bastankhah, Fernando Porté-Agel
    https://doi.org/10.1017/jfm.2016.595

    Attributes
    ----------
    model: Bastankhah2016Model
        The model for computing common data
    model_pars: dict
        Model parameters
    YAWM: float
        The yaw misalignment YAWM. If not given here
        it will be searched in the farm data.
    base_frame: foxes.core.WakeFrame
        The wake frame from which to start

    :group: models.wake_frames

    """

    def __init__(
        self,
        base_frame=RotorWD(),
        alpha=0.58,
        beta=0.07,
        induction="Madsen",
        **wake_k,
    ):
        """
        Constructor.

        Parameters
        ----------
        base_frame: foxes.core.WakeFrame
            The wake frame from which to start
        alpha: float
            model parameter used to determine onset of far wake region,
            if not found in wake model
        beta: float
            model parameter used to determine onset of far wake region,
            if not found in wake model
        induction: foxes.core.AxialInductionModel or str
            The induction model, if not found in wake model
        wake_k: dict, optional
            Parameters for the WakeK class, if not found in wake model

        """
        super().__init__()

        self.base_frame = base_frame
        self.model = None
        self.alpha = alpha
        self.beta = beta
        self.induction = induction
        self.wake_k = None
        self._wake_k_pars = wake_k

        setattr(self, FV.YAWM, 0.0)

    def __repr__(self):
        s = f"{type(self).__name__}("
        s += self.wake_k.repr() if self.wake_k is not None else ""
        s += ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return [self.wake_k, self.base_frame, self.model]

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        if not self.initialized:
            for w in algo.wake_models.values():
                if isinstance(w, Bastankhah2016):
                    if not w.initialized:
                        w.initialize(algo, verbosity, force)
                    self.model = w.model
                    self.wake_k = w.wake_k
                    break
            if self.model is None:
                self.model = Bastankhah2016Model(
                    alpha=self.alpha, beta=self.beta, induction=self.induction
                )
            if self.wake_k is None:
                wake_k = WakeK(**self._wake_k_pars)
                if not wake_k.all_none:
                    self.wake_k = wake_k
                else:
                    for w in algo.wake_models.values():
                        if hasattr(w, "wake_k"):
                            self.wake_k = w.wake_k
                            break

        super().initialize(algo, verbosity, force)

    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        return self.base_frame.calc_order(algo, mdata, fdata)

    def _update_y(self, algo, mdata, fdata, tdata, downwind_index, x, y):
        """
        Helper function for y deflection
        """

        # get gamma:
        gamma = self.get_data(
            FV.YAWM,
            FC.STATE_TARGET,
            lookup="wfs",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
            accept_nan=False,
        )
        gamma *= np.pi / 180

        # get k:
        k = self.wake_k(
            FC.STATE_TARGET,
            lookup_ti="f",
            lookup_k="sf",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=True,
            downwind_index=downwind_index,
            accept_nan=False,
        )

        # run model calculation:
        self.model.calc_data(algo, mdata, fdata, tdata, downwind_index, x, gamma, k)

        # select targets:
        st_sel = self.model.get_data(Bastankhah2016Model.ST_SEL, mdata)
        if np.any(st_sel):
            # prepare:
            n_st_sel = np.sum(st_sel)
            ydef = np.zeros((n_st_sel,), dtype=FC.DTYPE)

            # collect data:
            near = self.model.get_data(Bastankhah2016Model.NEAR, mdata)
            far = ~near

            # near wake:
            if np.any(near):
                # collect data:
                delta = self.model.get_data(Bastankhah2016Model.DELTA_NEAR, mdata)

                # set deflection:
                ydef[near] = delta

            # far wake:
            if np.any(far):
                # collect data:
                delta = self.model.get_data(Bastankhah2016Model.DELTA_FAR, mdata)

                # set deflection:
                ydef[far] = delta

            # apply deflection:
            y[st_sel] -= ydef

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        # get unyawed results:
        xyz = self.base_frame.get_wake_coos(
            algo,
            mdata,
            fdata,
            tdata,
            downwind_index,
        )

        # take rotor average:
        xy = np.einsum("stpd,p->std", xyz[..., :2], tdata[FC.TWEIGHTS])
        x = xy[:, :, 0]
        y = xy[:, :, 1]

        # apply deflection:
        self._update_y(algo, mdata, fdata, tdata, downwind_index, x, y)
        xyz[..., 1] = y[:, :, None]

        return xyz

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        points = self.base_frame.get_centreline_points(
            algo, mdata, fdata, downwind_index, x
        )
        tdata = TData.from_points(points)

        nx = np.zeros_like(points)
        nx[:, 0] = points[:, 1] - points[:, 0]
        nx[:, -1] = points[:, -1] - points[:, -2]
        nx[:, 1:-1] = 0.5 * (points[:, 1:-1] - points[:, :-2]) + 0.5 * (
            points[:, 2:] - points[:, 1:-1]
        )
        nx /= np.linalg.norm(nx, axis=-1)[:, :, None]

        nz = np.zeros_like(nx)
        nz[:, :, 2] = 1
        ny = np.cross(nz, nx, axis=-1)
        del nx, nz

        y = np.zeros_like(x)
        self._update_y(algo, mdata, fdata, tdata, downwind_index, x, y)

        points += y[:, :, None] * ny

        return points



from abc import abstractmethod
import numpy as np

from .data_calc_model import DataCalcModel
import foxes.constants as FC


class FarmDataModel(DataCalcModel):
    """
    Abstract base class for models that modify
    farm data.

    Attributes
    ----------
    pre_rotor: bool
        Flag for running this model before
        running the rotor model.

    :group: core

    """

    def __init__(self, pre_rotor=False):
        """
        Constructor.

        Parameters
        ----------
        pre_rotor: bool
            Flag for running this model before
            running the rotor model.

        """
        super().__init__()
        self.pre_rotor = pre_rotor

    @abstractmethod
    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return []

    def ensure_variables(self, algo, mdata, fdata):
        """
        Add variables to fdata, initialized with NaN

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        """
        n_states = fdata.n_states
        n_turbines = fdata.n_turbines
        for v in self.output_farm_vars(algo):
            if v not in fdata:
                fdata[v] = np.full((n_states, n_turbines), np.nan, dtype=FC.DTYPE)
                fdata.dims[v] = (FC.STATE, FC.TURBINE)

    @abstractmethod
    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        pass

    def run_calculation(self, algo, *data, out_vars, **calc_pars):
        """
        Starts the model calculation in parallel, via
        xarray's `apply_ufunc`.

        Typically this function is called by algorithms.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        *data: tuple of xarray.Dataset
            The input data
        out_vars: list of str
            The calculation output variables
        **calc_pars: dict, optional
            Additional arguments for the `calculate` function

        Returns
        -------
        results: xarray.Dataset
            The calculation results

        """
        return super().run_calculation(
            algo,
            *data,
            out_vars=out_vars,
            loop_dims=[FC.STATE],
            out_core_vars=[FC.TURBINE, FC.VARS],
            **calc_pars,
        )

    def __add__(self, m):
        if isinstance(m, list):
            return FarmDataModelList([self] + m)
        elif isinstance(m, FarmDataModelList):
            return FarmDataModelList([self] + m.models)
        else:
            return FarmDataModelList([self, m])


class FarmDataModelList(FarmDataModel):
    """
    A list of farm data models.

    By using the FarmDataModelList the models'
    `calculate` functions are called together
    under one common call of xarray's `apply_ufunc`.

    Attributes
    ----------
    models: list of foxes.core.FarmDataModel
        The model list

    :group: core

    """

    def __init__(self, models=[]):
        """
        Constructor.

        Parameters
        ----------
        models: list of foxes.core.FarmDataModel
            The model list

        """
        super().__init__()
        self.models = models

    def append(self, model):
        """
        Add a model to the list

        Parameters
        ----------
        model: foxes.core.FarmDataModel
            The model to add

        """
        self.models.append(model)

    def insert(self, index, model):
        """
        Insert a model into the list

        Parameters
        ----------
        index: int
            The index in the model list
        model: foxes.core.FarmDataModel
            The model to insert

        """
        self.models.insert(index, model)

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return self.models

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        ovars = []
        for m in self.models:
            ovars += m.output_farm_vars(algo)
        return list(dict.fromkeys(ovars))

    def calculate(self, algo, mdata, fdata, parameters=[]):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        parameters: list of dict, optional
            A list of parameter dicts, one for each model

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        if parameters is None:
            parameters = [{}] * len(self.models)
        elif not isinstance(parameters, list):
            raise ValueError(
                f"{self.name}: Wrong parameters type, expecting list, got {type(parameters).__name__}"
            )
        elif len(parameters) != len(self.models):
            raise ValueError(
                f"{self.name}: Wrong parameters length, expecting list with {len(self.models)} entries, got {len(parameters)}"
            )

        for mi, m in enumerate(self.models):
            # print("MLIST VARS BEFORE",m.name,list(fdata.keys()),parameters[mi])
            res = m.calculate(algo, mdata, fdata, **parameters[mi])
            fdata.update(res)

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



"""
Wind deficit wake models.
"""

from .jensen import JensenWake
from .bastankhah14 import Bastankhah2014
from .bastankhah16 import Bastankhah2016Model, Bastankhah2016
from .turbopark import TurbOParkWake, TurbOParkWakeIX



import numpy as np
import pandas as pd

from foxes.utils import Dict
from foxes.core import Turbine, TurbineType
import foxes.variables as FV


def read_turbine_type(wio_trbns, algo_dict, ws_exp_P, ws_exp_ct, verbosity):
    """
    Reads the turbine type from windio

    Parameters
    ----------
    wio_trbns: dict
        The windio turbines data
    algo_dict: dict
        The algorithm dictionary
    ws_exp_P: int
        The REWS exponent for power
    ws_exp_ct: int
        The REWS exponent for ct
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    ttype: str
        The turbine type model name

    :group: input.windio

    """
    tname = wio_trbns.pop("name")
    if verbosity > 2:
        print("    Reading wio_trbns")
        print("      Name:", tname)
        print("      Contents:", [k for k in wio_trbns.keys()])

    # read performance:
    performance = Dict(wio_trbns["performance"], name="performance")
    if verbosity > 2:
        print("        Reading performance")
        print("          Contents:", [k for k in performance.keys()])

    # P, ct data:
    if "power_curve" in performance:
        power_curve = Dict(performance["power_curve"], name="power_curve")
        if verbosity > 2:
            print("            Reading power_curve")
            print("              Contents:", [k for k in power_curve.keys()])
        P = power_curve["power_values"]
        ws_P = power_curve["power_wind_speeds"]
        ct_curve = Dict(performance["Ct_curve"], name="Ct_values")
        if verbosity > 2:
            print("            Reading Ct_curve")
            print("              Contents:", [k for k in ct_curve.keys()])
        ct = ct_curve["Ct_values"]
        ws_ct = ct_curve["Ct_wind_speeds"]

        data_P = pd.DataFrame(data={"ws": ws_P, "P": P})
        data_ct = pd.DataFrame(data={"ws": ws_ct, "ct": ct})

        def _get_wse_var(wse):
            if wse not in [1, 2, 3]:
                raise ValueError(f"Expecting wind speed exponent 1, 2 or 3, got {wse}")
            return FV.REWS if wse == 1 else (FV.REWS2 if wse == 2 else FV.REWS3)

        if verbosity > 2:
            print(f"            Creating model '{tname}'")
            print(f"              Turbine type class: PCtFromTwo")
        algo_dict["mbook"].turbine_types[tname] = TurbineType.new(
            ttype_type="PCtFromTwo",
            data_source_P=data_P,
            data_source_ct=data_ct,
            col_ws_P_file="ws",
            col_ws_ct_file="ws",
            col_P="P",
            col_ct="ct",
            H=wio_trbns["hub_height"],
            D=wio_trbns["rotor_diameter"],
            var_ws_ct=_get_wse_var(ws_exp_ct),
            var_ws_P=_get_wse_var(ws_exp_P),
            rho=1.225,
        )
        if verbosity > 2:
            print("               ", algo_dict["mbook"].turbine_types[tname])

    # P, ct data:
    elif "Cp_curve" in performance:
        cp_curve = Dict(performance["Cp_curve"], name="Cp_curve")
        if verbosity > 2:
            print("            Reading Cp_curve")
            print("              Contents:", [k for k in cp_curve.keys()])
        cp = cp_curve["Cp_values"]
        ws_cp = cp_curve["Cp_wind_speeds"]
        ct_curve = Dict(performance["Ct_curve"], name="Ct_values")
        if verbosity > 2:
            print("            Reading Ct_curve")
            print("              Contents:", [k for k in ct_curve.keys()])
        ct = ct_curve["Ct_values"]
        ws_ct = ct_curve["Ct_wind_speeds"]

        data_cp = pd.DataFrame(data={"ws": ws_cp, "cp": cp})
        data_ct = pd.DataFrame(data={"ws": ws_ct, "ct": ct})

        if verbosity > 2:
            print(f"            Creating model '{tname}'")
            print(f"              Turbine type class: CpCtFromTwo")
        algo_dict["mbook"].turbine_types[tname] = TurbineType.new(
            ttype_type="CpCtFromTwo",
            data_source_cp=data_cp,
            data_source_ct=data_ct,
            col_ws_cp_file="ws",
            col_ws_ct_file="ws",
            col_cp="cp",
            col_ct="ct",
            H=wio_trbns["hub_height"],
            D=wio_trbns["rotor_diameter"],
        )

    else:
        raise KeyError(f"Expecting either 'power_curve' or 'Cp_curve'")

    return tname


def read_layout(lname, ldict, algo_dict, ttype, verbosity=1):
    """
    Read wind farm layout from windio input

    Parameters
    ----------
    lname: str
        The layout name
    ldict: dict
        The layout data
    algo_dict: dict
        The algorithm dictionary
    ttype: str
        Name of the turbine type model
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    states: foxes.core.States
        The states object

    :group: input.windio

    """
    if verbosity > 2:
        print(f"        Reading '{lname}'")
    cdict = Dict(ldict["coordinates"], name="coordinates")
    farm = algo_dict["farm"]
    for xy in zip(cdict["x"], cdict["y"]):
        farm.add_turbine(
            Turbine(xy=np.array(xy), turbine_models=[ttype]),
            verbosity=verbosity-3,
        )
    if verbosity > 2:
        print(f"          Added {farm.n_turbines} wio_trbns of type '{ttype}'")



import numpy as np
import xarray as xr

from foxes.opt.core.farm_objective import FarmObjective
from foxes import variables as FV
import foxes.constants as FC


class FarmVarObjective(FarmObjective):
    """
    Objectives based on farm variables.

    Attributes
    ----------
    variable: str
        The variable name
    minimize: bool
        Switch for maximizing or minimizing
    deps: list of str
        The foxes variables on which the variable depends,
        or None for all
    rules: dict
        Contraction rules. Key: coordinate name str, value
        is str: min, max, sum, mean
    scale: float
        The scaling factor

    :group: opt.objectives

    """

    def __init__(
        self,
        problem,
        name,
        variable,
        contract_states,
        contract_turbines,
        minimize,
        deps=None,
        scale=1.0,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying optimization problem
        name: str
            The name of the objective function
        variable: str
            The foxes variable name
        contract_states: str
            Contraction rule for states: min, max, sum, mean
        contract_turbines: str
            Contraction rule for turbines: min, max, sum, mean
        minimize: bool
            Switch for maximizing or minimizing
        deps: list of str
            The foxes variables on which the variable depends,
            or None for all
        scale: float
            The scaling factor
        kwargs: dict, optional
            Additional parameters for `FarmObjective`

        """
        super().__init__(problem, name, **kwargs)
        self.variable = variable
        self.minimize = minimize
        self.deps = deps
        self.scale = scale
        self.rules = {FC.STATE: contract_states, FC.TURBINE: contract_turbines}

    def initialize(self, verbosity=0):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [not self.minimize]

    def vardeps_float(self):
        """
        Gets the dependencies of all components
        on the function float variables

        Returns
        -------
        deps: numpy.ndarray of bool
            The dependencies of components on function
            variables, shape: (n_components, n_vars_float)

        """
        if self.deps is None:
            return super().vardeps_float()

        out = np.zeros((self.n_components(), self.n_vars_float), dtype=bool)
        for i, tvr in enumerate(self.var_names_float):
            v, ti = self.problem.parse_tvar(tvr)
            if v in self.deps and ti in self.sel_turbines:
                out[0, i] = True

        return out

    def _contract(self, data):
        """
        Helper function for data contraction
        """
        for dim, rule in self.rules.items():
            if rule == "min":
                data = data.min(dim=dim)
            elif rule == "max":
                data = data.max(dim=dim)
            elif rule == "sum":
                data = data.sum(dim=dim)
            elif rule == "mean":
                data = data.mean(dim=dim)
            else:
                raise ValueError(
                    f"Objective '{self.name}': Unknown contraction for dimension '{dim}': '{rule}'. Choose: min, max, sum, mean"
                )
        return data

    def calc_individual(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_sel_components,)

        """
        data = problem_results[self.variable]
        if self.n_sel_turbines < self.farm.n_turbines:
            data = data[:, self.sel_turbines]
        data = self._contract(data) / self.scale

        return np.array([data], dtype=np.float64)

    def calc_population(self, vars_int, vars_float, problem_results, components=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results: Any
            The results of the variable application
            to the problem
        components: list of int, optional
            The selected components or None for all

        Returns
        -------
        values: np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = problem_results["n_pop"].values
        n_states = problem_results["n_org_states"].values
        n_turbines = problem_results.sizes[FC.TURBINE]
        data = (
            problem_results[self.variable]
            .to_numpy()
            .reshape(n_pop, n_states, n_turbines)
        )
        data = xr.DataArray(data, dims=(FC.POP, FC.STATE, FC.TURBINE))

        if self.n_sel_turbines < self.farm.n_turbines:
            data = data[:, self.sel_turbines]

        return self._contract(data / self.scale).to_numpy()[:, None]

    def finalize_individual(self, vars_int, vars_float, problem_results, verbosity=1):
        """
        Finalization, given the champion data.

        Parameters
        ----------
        vars_int: np.array
            The optimal integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The optimal float variable values, shape: (n_vars_float,)
        problem_results: Any
            The results of the variable application
            to the problem
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        values: np.array
            The component values, shape: (n_components,)

        """
        return (
            super().finalize_individual(
                vars_int, vars_float, problem_results, verbosity
            )
            * self.scale
        )


class MaxFarmPower(FarmVarObjective):
    """
    Maximize the mean wind farm power

    Parameters
    ----------
    problem: foxes.opt.FarmOptProblem
        The underlying optimization problem
    name: str
        The name of the objective function
    kwargs: dict, optional
        Additional parameters for `FarmVarObjective`

    :group: opt.objectives

    """

    def __init__(self, problem, name="maximize_power", **kwargs):
        if "scale" in kwargs:
            scale = kwargs.pop("scale")
        else:
            scale = 0.0
            ttypes = problem.algo.mbook.turbine_types
            for t in problem.farm.turbines:
                for mname in t.models:
                    if mname in ttypes:
                        scale += ttypes[mname].P_nominal
                        break

        super().__init__(
            problem,
            name,
            variable=FV.P,
            contract_states="mean",
            contract_turbines="sum",
            minimize=False,
            scale=scale,
            **kwargs,
        )


class MinimalMaxTI(FarmVarObjective):
    """
    Minimize the maximal turbine TI

    Parameters
    ----------
    problem: foxes.opt.FarmOptProblem
        The underlying optimization problem
    name: str
        The name of the objective function
    kwargs: dict, optional
        Additional parameters for `FarmVarObjective`

    :group: opt.objectives

    """

    def __init__(self, problem, name="minimize_TI", **kwargs):
        scale = kwargs.pop("scale") if "scale" in kwargs else 1.0
        super().__init__(
            problem,
            name,
            variable=FV.TI,
            contract_states="max",
            contract_turbines="max",
            minimize=True,
            scale=scale,
            **kwargs,
        )



from iwopy import Constraint

from foxes.utils import all_subclasses


class FarmConstraint(Constraint):
    """
    Abstract base class for foxes wind farm
    optimization constraints.

    :group: opt.core

    """

    def __init__(self, problem, name, sel_turbines=None, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying optimization problem
        name: str
            The name of the constraint
        sel_turbines: list of int, optional
            The selected turbines
        kwargs: dict, optional
            Additional parameters for `iwopy.Constraint`

        """
        super().__init__(problem, name, **kwargs)
        self._sel_turbines = sel_turbines

    @property
    def farm(self):
        """
        The wind farm

        Returns
        -------
        foxes.core.WindFarm :
            The wind farm

        """
        return self.problem.farm

    @property
    def sel_turbines(self):
        """
        The list of selected turbines

        Returns
        -------
        list of int :
            The list of selected turbines

        """
        return (
            self.problem.sel_turbines
            if self._sel_turbines is None
            else self._sel_turbines
        )

    @property
    def n_sel_turbines(self):
        """
        The numer of selected turbines

        Returns
        -------
        int :
            The numer of selected turbines

        """
        return len(self.sel_turbines)

    def add_to_layout_figure(self, ax, **kwargs):
        """
        Add to a layout figure

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The figure axis

        """
        return ax

    @classmethod
    def print_models(cls):
        """
        Prints all model names.
        """
        names = sorted([scls.__name__ for scls in all_subclasses(cls)])
        for n in names:
            print(n)



import numpy as np

from foxes.core import TurbineModel
import foxes.variables as FV
import foxes.constants as FC


class kTI(TurbineModel):
    """
    Calculates the wake model parameter `k`
    as a linear function of `TI`.

    Attributes
    ----------
    ti_var: str
        The `TI` variable name
    k_var: str
        The variable name for k

    :group: models.turbine_models

    """

    def __init__(self, kTI=None, kb=None, ti_var=FV.TI, ti_val=None, k_var=FV.K):
        """
        Constructor.

        Parameters
        ----------
        kTI: float, optional
            Uniform value for `kTI`. If not given it
            will be searched in farm data
        kb: float, optional
            Uniform value for `kb`. If not given it
            will be searched in farm data, and zero by default
        ti_var: str
            The `TI` variable name
        ti_val: float, optional
            The uniform value of `TI`. If not given it
            will be searched in farm data
        k_var: str
            The variable name for k

        """
        super().__init__()

        self.ti_var = ti_var
        self.k_var = k_var
        setattr(self, ti_var, ti_val)
        setattr(self, FV.KTI, kTI)
        setattr(self, FV.KB, 0 if kb is None else kb)

    def __repr__(self):
        kti = getattr(self, FV.KTI)
        kb = getattr(self, FV.KB)
        ti = getattr(self, self.ti_var)
        tiv = f", ti_val={ti}" if ti is not None else ""
        a = f"kTI={kti}, kb={kb}, ti_var={self.ti_var}{tiv}, k_var={self.k_var}"
        return f"{type(self).__name__}({a})"

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [self.k_var]

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        kti = self.get_data(
            FV.KTI, FC.STATE_TURBINE, lookup="sf", fdata=fdata, upcast=True
        )[st_sel]
        kb = self.get_data(
            FV.KB, FC.STATE_TURBINE, lookup="sf", fdata=fdata, upcast=True
        )[st_sel]
        ti = self.get_data(
            self.ti_var, FC.STATE_TURBINE, lookup="sf", fdata=fdata, upcast=True
        )[st_sel]

        k = fdata.get(
            self.k_var, np.zeros((fdata.n_states, fdata.n_turbines), dtype=FC.DTYPE)
        )

        k[st_sel] = kti * ti + kb

        return {self.k_var: k}



import numpy as np


def area(r1, r2, d):
    """
    Calculates the intersection of two
    circles with radii r1, r2 and a centre
    point distance d.

    Make sure that

        1) r1 >= r2
        2) r1 - r2 <= d <= r1 + r2

    These conditions are assumed but not checked
    within the function.

    Source:
    https://diego.assencio.com/?index=8d6ca3d82151bad815f78addf9b5c1c6

    Parameters
    ----------
    r1: float or numpy.ndarray
        The radius of circle 1
    r2: float or numpy.ndarray
        The radius of circle 2
    d: float or numpy.ndarray
        The distance between the centre points
        of the two circles

    Returns
    -------
    area: float or numpy.ndarray
        The intersectional area

    :group: utils.two_circles

    """
    d1 = (r1**2 - r2**2 + d**2) / (2 * d)
    d2 = d - d1

    a = np.maximum(np.minimum(d1 / r1, 1.0), -1)
    b = np.maximum(r1**2 - d1**2, 0.0)
    A1 = r1**2 * np.arccos(a) - d1 * np.sqrt(b)

    a = np.maximum(np.minimum(d2 / r2, 1.0), -1)
    b = np.maximum(r2**2 - d2**2, 0.0)
    A2 = r2**2 * np.arccos(a) - d2 * np.sqrt(b)

    """
    A1 = r1**2 * np.arccos(d1/r1) - d1 * np.sqrt(r1**2 - d1**2)
    A2 = r2**2 * np.arccos(d2/r2) - d2 * np.sqrt(r2**2 - d2**2)
    """

    return A1 + A2


def calc_area(r1, r2, d):
    """
    Calculates the intersection of two circles.

    All parameters should have the same shape,
    or be broadcastable to one another.

    Parameters
    ----------
    r1: numpy.ndarray
        The radius of circle 1
    r2: numpy.ndarray
        The radius of circle 2
    d: numpy.ndarray
        The distance between the centre points
        of the two circles

    Returns
    -------
    area: numpy.ndarray
        The intersectional area

    :group: utils.two_circles

    """

    # prepare:
    out = np.zeros_like(d)

    # condition d < r1 + r2:
    sel0 = d < r1 + r2
    if np.any(sel0):
        # condition r1 >= r2:
        sela = r1 >= r2
        selr = sel0 & sela
        if np.any(selr):
            # condition d <= r1 - r2:
            selb = d <= r1 - r2
            seld = selr & selb
            if np.any(seld):
                out[seld] = np.pi * r2[seld] ** 2

            # condition d > r1 - r2:
            seld = selr & (~selb)
            if np.any(seld):
                out[seld] = area(r1[seld], r2[seld], d[seld])

        # condition r1 < r2:
        selr = sel0 & (~sela)
        if np.any(selr):
            # condition d <= r2 - r1:
            selb = d <= r2 - r1
            seld = selr & selb
            if np.any(seld):
                out[seld] = np.pi * r1[seld] ** 2

            # condition d > r2 - r1:
            seld = selr & (~selb)
            if np.any(seld):
                out[seld] = area(r2[seld], r1[seld], d[seld])

    return out



import numpy as np


def logz(height, z0):
    """
    Calculates the log factor for
    wind speed profiles.

    Parameters
    ----------
    height: float
        The evaluation height
    z0: float
        The roughness length

    Returns
    -------
    lz: float
        The log factor

    :group: utils.abl.neutral

    """
    h = np.maximum(height, z0)
    return np.log(h / z0)


def ustar(ws_ref, h_ref, z0, kappa=0.41):
    """
    Calculates the friction velocity,
    based on reference data.

    Parameters
    ----------
    ws_ref: float
        The reference wind speed
    h_ref: float
        The reference height
    z0: float
        The roughness length
    kappa: float
        The van-Karman constant

    Returns
    -------
    ustar: float
        The friction velocity

    :group: utils.abl.neutral

    """
    lz = logz(h_ref, z0)
    return ws_ref * kappa / lz


def calc_ws(height, z0, ustar, kappa=0.41):
    """
    Calculate wind speeds at given height

    Parameters
    ----------
    height: float
        The evaluation height
    z0: float
        The roughness length
    ustar: float
        The friction velocity
    kappa: float
        The van-Karman constant

    Returns
    -------
    ws: float
        The wind speed

    :group: utils.abl.neutral

    """
    return ustar / kappa * logz(height, z0)
